from typing import BinaryIO
from struct import unpack


# https://github.com/OpenL2D/moc3ingbird/blob/master/src/moc3.hexpat
def read_moc3(file: BinaryIO):
    """Reads a MOC3 file and returns Part names and Parameter names

    Args:
        file (BinaryIO): File pointer

    Returns:
        tuple: (list[str], list[str]), Part names and Parameter names
    """
    # Header: 64 bytes
    file.seek(0)
    assert file.read(4) == b"MOC3"
    version = unpack("<c", file.read(1))[0]
    isBigEndian = unpack("<b", file.read(1))[0]
    assert not isBigEndian

    # TODO: Other fields
    file.seek(0x40)
    pCountInfo = unpack("<I", file.read(4))[0]

    file.seek(pCountInfo)
    numParts = unpack("<I", file.read(4))[0]
    file.seek(0x10, 1)
    numParameters = unpack("<I", file.read(4))[0]

    file.seek(0x4C)
    pParts = unpack("<I", file.read(4))[0]

    file.seek(0x108)
    pParameters = unpack("<I", file.read(4))[0]

    def read_strings(offset, count):
        for i in range(0, count):
            file.seek(offset + i * 0x40)
            buffer = bytearray()
            while b := file.read(1)[0]:
                buffer.append(b)
            yield buffer.decode(encoding="utf-8")

    return list(read_strings(pParts, numParts)), list(
        read_strings(pParameters, numParameters)
    )



from logging import getLogger
from sssekai.unity.AssetBundle import load_assetbundle
from UnityPy.enums import ClassIDType
from sssekai.fmt.rla import read_archive_rla_frames, read_rla_frames
import os, io, json, base64

logger = getLogger(__name__)


def main_rla2json(args):
    def __dump_from_frame_gen(frame_gen):
        outdir = args.outdir
        os.makedirs(outdir, exist_ok=True)
        ensure_dir = lambda x: os.makedirs(os.path.dirname(x), exist_ok=True) or x
        for tick, frame in frame_gen:
            print("tick: %16s, type: %32s" % (tick, frame["type"]), end="\r")
            match frame["type"]:
                case "SoundData":
                    if frame["encoding"] == "hca":
                        with open(
                            ensure_dir(
                                os.path.join(outdir, "SoundData", str(tick) + ".hca"),
                            ),
                            "wb",
                        ) as f:
                            raw_data = frame["data"]
                            start = raw_data.find(b"\xFF\xFF")
                            f.write(raw_data[start:])
                    else:
                        raise ValueError("unsupported encoding: %s" % frame["encoding"])
                case _:
                    with open(
                        ensure_dir(
                            os.path.join(outdir, frame["type"], str(tick) + ".json")
                        ),
                        "w",
                        encoding="utf-8",
                    ) as f:
                        json.dump(frame, f, ensure_ascii=False, indent=4)

    if os.path.isfile(args.input):
        with open(args.input, "rb") as f:
            datas = dict()
            rla_env = load_assetbundle(f)
            for obj in rla_env.objects:
                if obj.type in {ClassIDType.TextAsset}:
                    data = obj.read()
                    datas[data.m_Name] = data.m_Script.encode(
                        "utf-8", "surrogateescape"
                    )
            header = datas.get("sekai.rlh", None)
            assert header, "RLH Header file not found!"
            version = tuple(map(int, header["version"].split(".")))
            logger.info("Version: %d.%d" % version)
            logger.info("Count: %d" % len(header["splitFileIds"]))
            splitSeconds = header["splitSeconds"]
            for sid in header["splitFileIds"]:
                sname = "sekai_%02d_%08d" % (splitSeconds, sid)
                script = datas[sname + ".rla"]
                frame_gen = read_archive_rla_frames(
                    io.BytesIO(script), version, args.strict
                )
                __dump_from_frame_gen(frame_gen)
    else:
        logger.info("Reading from directory")
        packet_gen = (
            (file, open(os.path.join(args.input, file), "rb").read())
            for file in os.listdir(args.input)
        )
        version = tuple(map(int, args.version.split(".")))
        frame_gen = read_rla_frames(packet_gen, version, args.strict)
        __dump_from_frame_gen(frame_gen)



import math, fsspec
from typing import Callable
from functools import cached_property, cache
from collections import defaultdict
from fsspec.spec import AbstractBufferedFile
from fsspec.caching import BaseCache, register_cache
from fsspec.archive import AbstractArchiveFileSystem
from requests import Response
from logging import getLogger
from sssekai.crypto.AssetBundle import decrypt_iter, SEKAI_AB_MAGIC
from . import AbCache, AbCacheEntry

logger = getLogger("abcache.fs")


# Reference: https://github.com/fsspec/filesystem_spec/blame/master/fsspec/caching.py
class UnidirectionalBlockCache(BaseCache):
    """Block-based cache that only fetches data in one direction with a fixed block size"""

    name: str = "unidirectional_blockcache"

    def __init__(
        self,
        blocksize: int,
        fetcher: Callable[[int, int], bytes],
        size: int,
        ignore_size: bool = False,
    ) -> None:
        """Create a unidirectional block cache.

        Args:
            blocksize (int): Block size in bytes.
            fetcher (Callable[[int, int], bytes]): Fetcher function that takes start and end byte positions and returns the data.
            size (int): Size of the file.
            ignore_size (bool, optional): Don't truncate reads with `size` provided. Defaults to False.
        """
        super().__init__(blocksize, fetcher, size)
        self.ignore_size = ignore_size
        if not ignore_size:
            self.nblocks = math.ceil(size / self.blocksize)
        else:
            self.nblocks = float("inf")
        self.blocks = list()
        self.eof = False

    def __fetch_block(self, block_number):
        assert block_number < self.nblocks, "block out of range"
        while not self.eof and len(self.blocks) - 1 < block_number:
            start = self.blocksize * len(self.blocks)
            end = start + self.blocksize
            block = self.fetcher(start, end)
            if block:
                self.blocks.append(block)
            else:
                self.eof = True
        if block_number < len(self.blocks):
            return self.blocks[block_number]
        return b""  # EOF behavior when ignore_size is True

    def _fetch(self, start: int | None, stop: int | None) -> bytes:
        if start is None:
            start = 0
        if stop is None:
            stop = self.size
        if not self.ignore_size:
            stop = min(stop, self.size)  # XXX: why didn't fsspec handle this?
        if (not self.ignore_size and start >= self.size) or start >= stop:
            return b""
        start_blk, start_pos = start // self.blocksize, start % self.blocksize
        end_blk, end_pos = stop // self.blocksize, stop % self.blocksize

        if start_blk == end_blk:
            out = self.__fetch_block(start_blk)[start_pos:end_pos]
            return out
        else:
            out = [self.__fetch_block(start_blk)[start_pos:]]
            out += [self.__fetch_block(blk) for blk in range(start_blk + 1, end_blk)]
            out += [self.__fetch_block(end_blk)[:end_pos]]
            return b"".join(out)


register_cache(UnidirectionalBlockCache)


# Reference: https://github.com/fsspec/filesystem_spec/blob/master/fsspec/implementations/http.py#L526
class AbCacheFile(AbstractBufferedFile):
    """Cached, file-like object for reading from an AbCache on demand.

    Note:
        - The fetched content is decrypted on the fly.
        - Seeks are simulated by read-aheads (by UnidirectionalBlockCache). Meaning seek operations
          will incur additional download (in-betweens will be cached as well).
        - File sizes reported are *inaccurate* due to wrong values sent by the server.
          Read until EOF otherwise you will miss data.
    """

    DEFAULT_BLOCK_SIZE = 65536  # 64KB
    entry: AbCacheEntry

    @property
    def session(self) -> AbCache:
        return self.fs.cache

    @property
    def entry(self) -> AbCacheEntry:
        entry = self.session.get_entry_by_bundle_name(self.path.strip("/"))
        assert entry is not None, "entry not found"
        return entry

    def read(self, length=-1):
        if length < 0:
            length = float("inf")
        out = self.cache._fetch(self.loc, self.loc + length)
        self.loc += len(out)
        return out

    def __init__(self, fs, bundle: str, block_size=None):
        self.fs, self.path = fs, bundle
        self.fetch_loc = 0
        super().__init__(
            fs,
            bundle,
            block_size=block_size or self.DEFAULT_BLOCK_SIZE,
            mode="rb",
            cache_type="unidirectional_blockcache",
            size=self.entry.fileSize,
            cache_options={"ignore_size": True},
            # Sadly entry size could be *extremely* inaccurate.
            # We have to ignore it and fetch until EOF.
        )

    @cached_property
    def __resp(self) -> Response:
        url = self.session.get_entry_download_url(self.entry)
        resp = self.session.get(url, stream=True)
        resp.raise_for_status()
        return resp

    @cached_property
    def __fetch(self):
        def __innner():
            for block in decrypt_iter(
                lambda nbytes: next(self.__resp.iter_content(nbytes), b""),
                self.blocksize,
            ):
                yield bytes(block)

        return __innner()

    def _fetch_range(self, start, end):
        assert start - self.fetch_loc == 0, "can only fetch sequentially"
        self.fetch_loc = end
        return next(self.__fetch, b"")


# Reference: https://github.com/fsspec/filesystem_spec/blob/master/fsspec/implementations/libarchive.py
class AbCacheFilesystem(AbstractArchiveFileSystem):
    """Filesystem for reading from an AbCache on demand."""

    root_marker = "/"
    protocol = "abcache"
    cache: AbCache

    def __init__(self, fo: str = "", cache_obj: AbCache = None, *args, **kwargs):
        """Initialize the filesystem with a cache object
        or a file-like object that contains the cache database file.

        Args:
            fo (str, optional): the cahce database file object . Defaults to "".
            cache_obj (AbCache, optional): the cache database. Defaults to None.
        """
        if cache_obj:
            self.cache = cache_obj
        else:
            self.cache = AbCache()
            if isinstance(fo, str):
                with fsspec.open(fo, "rb") as f:
                    self.cache.load(f)
            else:
                self.cache.load(fo)
            self.cache.update_client_headers()

    @cached_property
    def dir_cache(self):
        # Reference implementation did O(n) per *every* ls() call
        # We can make it O(1) with DP on tree preprocessing of O(nlogn)
        bundles = self.cache.abcache_index.bundles
        # Only the leaf nodes are given.
        keys = set((self.root_marker + key for key in bundles.keys()))
        keys |= self._all_dirnames(bundles.keys())
        # Sorting implies DFS order.
        keys = [self.root_marker] + [key for key in sorted(keys)]
        _trim = lambda key: key[len(self.root_marker) :]
        nodes = [
            {
                "name": key,
                "type": "directory" if not _trim(key) in bundles else "file",
                "size": (
                    0 if not _trim(key) in bundles else bundles[_trim(key)].fileSize
                ),
                "item_count": 0,
                "file_count": 0,
                "total_size": 0,
            }
            for key in keys
        ]
        # Already in DFS order.
        # Get start index for each directory and their item count.
        stack = [0]
        graph = defaultdict(list)
        table = {node["name"]: index for index, node in enumerate(nodes)}

        def is_file(name):
            return _trim(name) in bundles

        def is_parent_path(a, b):
            # a is parent of b
            if a == self.root_marker:
                return True
            return b.startswith(a + self.root_marker)

        def maintain():
            # Always starts from root. Safe to assume stack size >= 2
            u, v = stack[-2], stack[-1]
            nodes[u]["item_count"] += nodes[v]["item_count"]
            nodes[u]["file_count"] += nodes[v]["file_count"]
            nodes[u]["total_size"] += nodes[v]["total_size"]
            stack.pop()

        for index, name in enumerate(keys):
            # Skip root
            if index == 0:
                continue
            while not is_parent_path(keys[stack[-1]], name):
                maintain()
            pa = stack[-1]
            nodes[pa]["item_count"] += 1
            graph[pa].append(index)
            if not is_file(name):
                stack.append(index)
            else:
                nodes[pa]["file_count"] += 1
                nodes[pa]["total_size"] += nodes[index]["size"]
                nodes[index]["total_size"] = nodes[index]["size"]
        while len(stack) >= 2:
            maintain()
        assert nodes[0]["file_count"] == len(bundles), "file count mismatch"
        return nodes, graph, table

    def _get_dirs(self):
        return self.dir_cache

    def info(self, path, **kwargs):
        nodes, graph, table = self._get_dirs()
        path = path or self.root_marker
        if path in table:
            return nodes[table[path]]
        else:
            raise FileNotFoundError(path)

    @cache
    def ls(self, path, detail=True, **kwargs):
        nodes, graph, table = self._get_dirs()
        path = path or self.root_marker
        if path in table:
            u = table[path]
            return [nodes[v] if detail else nodes[v]["name"] for v in graph[u]]
        return []

    def open(self, path, mode="rb"):
        assert mode == "rb", "only binary read-only mode is supported"
        return AbCacheFile(self, path)


fsspec.register_implementation("abcache", AbCacheFilesystem)



import zipfile
import UnityPy
import logging
import re

from io import BytesIO

import UnityPy.classes
import UnityPy.enums
import UnityPy.enums.ClassIDType
from sssekai.unity.AssetBundle import load_assetbundle
from tqdm import tqdm

HASHREGEX = re.compile(b"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}")
REGION_MAP = {
    "com.sega.pjsekai": "jp",
    "com.sega.ColorfulStage.en": "en",
    "com.hermes.mk.asia": "tw",
    "com.pjsekai.kr": "kr",
    "com.hermes.mk": "cn",
}
logger = logging.getLogger("apphash")


def enum_candidates(zip_file, filter):
    return (
        (f, zip_file.open(f), zip_file) for f in zip_file.filelist if filter(f.filename)
    )


def enum_package(zip_file):
    yield zip_file
    for f in zip_file.filelist:
        if f.filename.lower().endswith(".apk"):
            yield zipfile.ZipFile(zip_file.open(f))


def main_apphash(args):
    env = UnityPy.Environment()
    app_package = "unknown"
    app_version = "unknown"
    app_hash = "unknown"

    if not args.ab_src:
        if not args.apk_src or args.fetch:
            from requests import get

            src = BytesIO()
            logger.debug("Fetching latest game package (JP) from APKPure")
            resp = get(
                "https://d.apkpure.net/b/XAPK/com.sega.pjsekai?version=latest",
                stream=True,
            )
            size = resp.headers.get("Content-Length", -1)
            with tqdm(total=int(size), unit="B", unit_scale=True) as progress:
                progress.sp = print
                for chunck in resp.iter_content(chunk_size=2**20):
                    src.write(chunck)
                    progress.update(len(chunck))
            src.seek(0)
        else:
            src = open(args.apk_src, "rb")
        with zipfile.ZipFile(src, "r") as zip_ref:
            manifests = [
                manifest
                for package in enum_package(zip_ref)
                for manifest in enum_candidates(
                    package, lambda fn: fn == "AndroidManifest.xml"
                )
            ]
            manifest = manifests[0][1]
            from pyaxmlparser.axmlprinter import AXMLPrinter

            manifest = AXMLPrinter(manifest.read()).get_xml_obj()
            find_key = lambda ky: next((k for k in manifest.keys() if ky in k), None)
            app_version = manifest.get(find_key("versionName"), None)
            app_package = manifest.get(find_key("package"), None)
            logger.info("Package: %s" % app_package)
            logger.info("Version: %s" % app_version)
            candidates = [
                candidate
                for package in enum_package(zip_ref)
                for candidate in enum_candidates(
                    package,
                    lambda fn: fn.split("/")[-1]
                    in {
                        "6350e2ec327334c8a9b7f494f344a761",  # PJSK Android
                        "c726e51b6fe37463685916a1687158dd",  # PJSK iOS
                        "data.unity3d",  # TW,KR (ByteDance)
                    },
                )
            ]
            for candidate, stream, _ in candidates:
                env.load_file(stream)
    else:
        logger.info("Loading from AssetBundle %s" % args.ab_src)
        with open(args.ab_src, "rb") as f:
            env = load_assetbundle(BytesIO(f.read()))
    for pobj in env.objects:
        # TODO: Dump actual typetree data from the game itself?
        if pobj.type == UnityPy.enums.ClassIDType.MonoBehaviour:
            obj = pobj.read(check_read=False)
            for name in {"production_android", "production_ios"}:
                if obj.m_Name == name:
                    hashStr = HASHREGEX.finditer(pobj.get_raw_data())
                    for m in hashStr:
                        ans = m.group().decode()
                        logger.debug("%s: %s" % (name, ans))
                        app_hash = ans
    region = REGION_MAP.get(app_package, app_package)
    print(app_version, region, app_hash)



import codecs
import sys, os
import argparse
from sssekai.entrypoint.apidecrypt import main_apidecrypt
from sssekai.entrypoint.abdecrypt import main_abdecrypt
from sssekai.entrypoint.rla2json import main_rla2json
from sssekai.entrypoint.usmdemux import main_usmdemux
from sssekai.entrypoint.abcache import main_abcache, DEFAULT_CACHE_DB_FILE
from sssekai.entrypoint.abserve import main_abserve
from sssekai.entrypoint.live2dextract import main_live2dextract
from sssekai.entrypoint.spineextract import main_spineextract
from sssekai.entrypoint.apphash import main_apphash
from sssekai.entrypoint.mvdata import main_mvdata
from sssekai.entrypoint.moc3paths import main_moc3paths
from sssekai.unity import sssekai_get_unity_version, sssekai_set_unity_version


def create_parser(clazz=argparse.ArgumentParser):
    parser = clazz(
        description="""Project SEKAI Asset Utility / PJSK 资源工具""",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "--log-level",
        type=str,
        help="logging level (default: %(default)s)",
        default="INFO",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
    )
    parser.add_argument(
        "--unity-version",
        type=str,
        help="""Unity version to use (default: %(default)s)
Prior to game version 3.6.0 (JP), this has always been 2020.3.21f1
This has been changed to 2022.3.21f1 since, which would apply to all the assets from 3.6.0 onwards.
If you encounter any issues, try switching to the old version, or vice versa.""",
        default=sssekai_get_unity_version(),
    )
    subparsers = parser.add_subparsers(
        title="subcommands", description="valid subcommands", help="additional help"
    )
    # apidecrypt
    apidecrypt_parser = subparsers.add_parser(
        "apidecrypt",
        usage="""API crypto dumper
This crypto applies to:
    - API request/response body dumped by packet sniffer (mitmproxy, wireshark, etc.)
    - AssetBundleInfo (see sssekai.abcache)""",
    )
    apidecrypt_parser.add_argument("infile", type=str, help="input dump file")
    apidecrypt_parser.add_argument("outfile", type=str, help="output json file")
    apidecrypt_parser.add_argument(
        "--region",
        type=str,
        help="app region",
        default="jp",
        choices=["jp", "tw", "en", "kr", "cn"],
    )
    apidecrypt_parser.set_defaults(func=main_apidecrypt)
    # abdecrypt
    abdecrypt_parser = subparsers.add_parser(
        "abdecrypt", usage="""Decrypt Sekai AssetBundle"""
    )
    abdecrypt_parser.add_argument("indir", type=str, help="input directory")
    abdecrypt_parser.add_argument("outdir", type=str, help="output directory")
    abdecrypt_parser.set_defaults(func=main_abdecrypt)
    # usmdemux
    usmdemux_parser = subparsers.add_parser(
        "usmdemux", usage="""Demux Sekai USM Video in a AssetBundle"""
    )
    usmdemux_parser.add_argument("infile", type=str, help="input file")
    usmdemux_parser.add_argument("outdir", type=str, help="output directory")
    usmdemux_parser.set_defaults(func=main_usmdemux)
    # abcache
    abcache_parser = subparsers.add_parser(
        "abcache", usage="""Sekai AssetBundle Metadata Cache / Game API Helper"""
    )
    group = abcache_parser.add_argument_group("save/load options")
    group.add_argument(
        "--db",
        type=str,
        help="""cache database file path (default: %(default)s)""",
        default=DEFAULT_CACHE_DB_FILE,
    )
    group.add_argument(
        "--no-keep-auth",
        action="store_false",
        help="don't keep authentication data in the cache database during the update",
    )
    group.add_argument(
        "--no-update",
        action="store_true",
        help="skip all metadata updates and use cached ones as is. this supercedes 'game version options' and 'authentication options'",
    )
    group = abcache_parser.add_argument_group(
        "game version options",
        "NOTE: these options are used to *update* the cache database. Use --no-update to skip updating. Also, you may want to read the Wiki to know how to get these values.",
    )
    group.add_argument(
        "--app-region",
        type=str,
        help="PJSK app region (default: %(default)s)",
        choices=["jp", "tw", "en", "kr", "cn"],
        default="jp",
    )
    group.add_argument(
        "--app-platform",
        type=str,
        help="PJSK app platform (default: %(default)s",
        choices=["ios", "android"],
        default="android",
    )  # Let's hope this doesn't age well...
    group.add_argument(
        "--app-version",
        type=str,
        help="PJSK app version. This is required unless --no-update is specified",
        default="",
    )
    group.add_argument(
        "--app-appHash",
        type=str,
        help="PJSK app hash. This is required unless --no-update is specified",
        default="",
    )
    group.add_argument(
        "--app-abVersion",
        type=str,
        help="PJSK AssetBundle URL version. This is used for ROW servers since it may differ.",
        default=None,
    )
    group = abcache_parser.add_argument_group("download options")
    group.add_argument(
        "--download-filter",
        type=str,
        help="filter AssetBundles (by bundle names) with this regex pattern",
        default=None,
    )
    group.add_argument(
        "--download-dir",
        type=str,
        help="asset bundle download directory. leave empty if you don't want to download anything",
        default="",
    )
    group.add_argument(
        "--download-ensure-deps",
        action="store_true",
        help="ensure dependencies (of the downloaded ones) are downloaded as well",
    )
    group.add_argument(
        "--download-workers",
        type=int,
        help="number of download workers (default: %(default)s)",
        default=4,
    )
    group = abcache_parser.add_argument_group(
        "extra options",
        "NOTE: when *any* of these options are specified, the cache database *won't* be updated, and no download will be performed either.",
    )
    group.add_argument(
        "--dump-master-data",
        type=str,
        help="directory to store the dumped master data in JSON.",
        default=None,
    )
    group.add_argument(
        "--dump-user-data",
        type=str,
        help="directory to store the dumped master data of the authencicated user in JSON. NOTE: require proper authentication required",
        default=None,
    )
    group.add_argument(
        "--keep-compact",
        help="keep compacted datasets as is without expanding them to key-value pairs",
        action="store_true",
    )
    group = abcache_parser.add_argument_group(
        "authentication arguments",
        "Only needed for some functionailties (i.e. --dump-master-data)",
    )
    group.add_argument(
        "--auth-credential",
        type=str,
        help="Credential for API authentication. Can be omitted on JP/EN servers, in which case a guest account would be automatically registered and used.",
        default=None,
    )
    abcache_parser.set_defaults(func=main_abcache)
    abserve_parser = subparsers.add_parser(
        "abserve", usage="""AbCache Filesystem Server / Backend"""
    )
    abserve_parser.add_argument(
        "--db",
        type=str,
        help="""cache database file path (default: %(default)s)""",
        default=DEFAULT_CACHE_DB_FILE,
    )
    abserve_parser.add_argument(
        "--fuse",
        type=str,
        help="""local mount point for AbCache. Required FUSE on host OS and fusepy (default: %(default)s)""",
        default="",
    )
    abserve_parser.add_argument(
        "--host",
        type=str,
        help="""address of the interface to listen on (default: %(default)s)""",
        default="0.0.0.0",
    )
    abserve_parser.add_argument(
        "--port",
        type=int,
        help="""port to listen on (default: %(default)s)""",
        default="3939",
    )
    abserve_parser.set_defaults(func=main_abserve)
    # live2dextract
    live2dextract_parser = subparsers.add_parser(
        "live2dextract", usage="""Extract Sekai Live2D Models in a AssetBundle"""
    )
    live2dextract_parser.add_argument("infile", type=str, help="input file")
    live2dextract_parser.add_argument("outdir", type=str, help="output directory")
    live2dextract_parser.add_argument(
        "--no-anim", action="store_true", help="don't extract animation clips"
    )
    live2dextract_parser.set_defaults(func=main_live2dextract)
    # spineextract
    spineextract_parser = subparsers.add_parser(
        "spineextract",
        usage="""Extract Sekai Spine (Esoteric Spine2D) Models in a AssetBundle""",
    )
    spineextract_parser.add_argument("infile", type=str, help="input file")
    spineextract_parser.add_argument("outdir", type=str, help="output directory")
    spineextract_parser.set_defaults(func=main_spineextract)
    # rla2json
    rla2json_parser = subparsers.add_parser(
        "rla2json",
        usage="""Read streaming_live/archive files and dump their information to JSON or HCA packets""",
    )
    rla2json_parser.add_argument(
        "input",
        type=str,
        help="input archive file or directory containing loose packets",
    )
    rla2json_parser.add_argument("outdir", type=str, help="output directory")
    rla2json_parser.add_argument(
        "--version",
        type=str,
        help="RLA version. Only used for loose packets",
        default=(1, 6),
        choices=["1.%1s" % x for x in range(0, 6 + 1)],
    )
    rla2json_parser.add_argument(
        "--strict",
        action="store_true",
        help="strict mode. raise exception on unknown frame type or corrupt frames",
    )
    rla2json_parser.set_defaults(func=main_rla2json)
    # apphash
    apphash_parser = subparsers.add_parser(
        "apphash",
        usage="""Download/extract game AppHash values
*NOTE*: Only [app version] [app region] [app hash] are written to stdout if successful. Fields are separated by space. Unavailable fields are left as 'unknown'.""",
    )
    apphash_parser.add_argument(
        "--apk-src",
        type=str,
        help="APK source file (default: fetch from APKPure)",
        default=None,
    )
    apphash_parser.add_argument(
        "--ab-src",
        type=str,
        help="AssetBundle (namely, 6350e2ec327334c8a9b7f494f344a761 or c726e51b6fe37463685916a1687158dd) source file (default: fetch from APK)",
        default=None,
    )
    apphash_parser.add_argument(
        "--fetch", action="store_true", help="force fetching the latest APK"
    )
    apphash_parser.set_defaults(func=main_apphash)
    # mvdata
    mvdata_parser = subparsers.add_parser(
        "mvdata", usage="""Extract MV Data from AssetBundle"""
    )
    mvdata_parser.add_argument(
        "infile", type=str, help="cache directory (live_pv/mv_data)"
    )
    mvdata_parser.add_argument("outdir", type=str, help="output JSON file to dump into")
    mvdata_parser.set_defaults(func=main_mvdata)
    # moc3paths
    moc3paths_parser = subparsers.add_parser(
        "moc3paths", usage="""Extract animation path CRCs from raw .moc3 binaries"""
    )
    moc3paths_parser.add_argument("indir", type=str, help="input directory")
    moc3paths_parser.set_defaults(func=main_moc3paths)
    return parser


def __main__():
    from tqdm.std import tqdm as tqdm_c

    class SemaphoreStdout:
        @staticmethod
        def write(__s):
            # Blocks tqdm's output until write on this stream is done
            # Solves cases where progress bars gets re-rendered when logs
            # spews out too fast
            with tqdm_c.external_write_mode(file=sys.stdout, nolock=False):
                return sys.stdout.write(__s)

    # parse args
    parser = create_parser(argparse.ArgumentParser)
    args = parser.parse_args()
    # set logging level
    import coloredlogs
    from logging import basicConfig

    coloredlogs.install(
        level=args.log_level,
        format="%(asctime)s | %(levelname).1s | %(name)s %(message)s",
        datefmt="%H:%M:%S",
        isatty=True,
        stream=SemaphoreStdout,
    )
    basicConfig(
        level=args.log_level,
        format="%(asctime)s | %(levelname).1s | %(name)s %(message)s",
        datefmt="%H:%M:%S",
        stream=SemaphoreStdout,
    )
    # override unity version
    sssekai_set_unity_version(args.unity_version)
    if "func" in args:
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    __main__()
    sys.exit(0)



import sys, os

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)))
import setuptools, sssekai

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sssekai",
    version=sssekai.__version__,
    author="greats3an",
    author_email="greats3an@gmail.com",
    description="Project SEKAI Asset Utility / PJSK 资源工具",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mos9527/sssekai",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "msgpack",
        "pycryptodome",
        "unitypy >= 1.20.25",
        "astc-encoder-py >= 0.1.11",  # https://github.com/mos9527/sssekai_blender_io/issues/11
        "wannacri",
        "python-json-logger",
        "tqdm",
        "rich",
        "coloredlogs",
        "requests",
        "pyaxmlparser",
    ],
    extras_require={
        "gui": ["Gooey @ git+https://github.com/nicolasbraun/Gooey@11bd412"]
    },
    entry_points={
        "console_scripts": [
            "sssekai = sssekai.__main__:__main__",
            "sssekai-gui = sssekai.__gui__:__main__",
        ],
        "fsspec.specs": ["abcache = sssekai.abcache.fs.AbCacheFilesystem"],
    },
    python_requires=">=3.10",
)
